"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const dompurify = require("dompurify");
const hljs = require("highlight.js");
const marked = require("marked");
const React = require("react");
const generateScrollingAnchorId = (line) => {
    return "scrolling-anchor-id-" + line;
};
class MarkdownPreview extends React.PureComponent {
    constructor(props) {
        super(props);
        this._subscriptions = [];
        const colors = {
            background: this.props.oni.colors.getColor("editor.background"),
            foreground: this.props.oni.colors.getColor("editor.foreground"),
            link: this.props.oni.colors.getColor("highlight.mode.normal.background"),
            codeBackground: this.props.oni.colors.getColor("background"),
            codeForeground: this.props.oni.colors.getColor("foreground"),
            codeBorder: this.props.oni.colors.getColor("toolTip.border"),
        };
        this.state = { source: "", colors };
    }
    componentDidMount() {
        const activeEditor = this.props.oni.editors.activeEditor;
        if (!activeEditor) {
            return;
        }
        this.subscribe(activeEditor.onBufferChanged, args => this.onBufferChanged(args));
        // TODO: Subscribe "onFocusChanged"
        if (this.props.oni.configuration.getValue("experimental.markdownPreview.autoScroll")) {
            this.subscribe(activeEditor.onBufferScrolled, args => this.onBufferScrolled(args));
        }
        this.previewBuffer(activeEditor.activeBuffer);
    }
    componentWillUnmount() {
        for (const subscription of this._subscriptions) {
            subscription.dispose();
        }
        this._subscriptions = [];
    }
    render() {
        const renderedMarkdown = this.generateMarkdown();
        this.props.instance.updateContent(this.state.source, renderedMarkdown);
        const html = renderedMarkdown + this.generateContainerStyle();
        const classes = "stack enable-mouse oniPluginMarkdownPreviewContainerStyle";
        return React.createElement("div", { className: classes, dangerouslySetInnerHTML: { __html: html } });
    }
    generateContainerStyle() {
        const colors = this.state.colors;
        const syntaxHighlightTheme = this.props.oni.configuration.getValue("experimental.markdownPreview.syntaxTheme");
        const codeBlockStyle = `
            background: ${colors.codeBackground};
            color: ${colors.codeForeground};
            border-color: ${colors.codeBackground};
            font-family: Consolas,Menlo,Monaco,Lucida Console,Liberation Mono,DejaVu Sans Mono,Bitstream Vera Sans Mono,Courier New,monospace,sans-serif
            padding: 0.4em 0.4em 0.4em 0.4em;
            margin: 0.4em 0.4em 0.4em 0.4em;
        `;
        return `
            <link rel="stylesheet" href="node_modules/highlight.js/styles/${syntaxHighlightTheme}.css">

            <style>
            .oniPluginMarkdownPreviewContainerStyle {
                padding: 1em 1em 1em 1em;
                overflow-y: auto;
                background: ${colors.background};
                color: ${colors.foreground};
            }

            .oniPluginMarkdownPreviewContainerStyle a:link {
                color: ${colors.link};
            }

            .oniPluginMarkdownPreviewContainerStyle pre {
                display: flex;
                ${codeBlockStyle}
            }

            .oniPluginMarkdownPreviewContainerStyle code {
                ${codeBlockStyle}
            }
            </style>
        `;
    }
    generateMarkdown() {
        const markdownLines = dompurify.sanitize(this.state.source).split("\n");
        const generateAnchor = (line) => {
            return `<a id="${generateScrollingAnchorId(line)}"></a>`;
        };
        let isBlock = false;
        const originalLinesCount = markdownLines.length - 1;
        // tslint:disable-next-line
        for (var i = originalLinesCount; i > 0; i--) {
            if (markdownLines[i].includes("```")) {
                isBlock = !isBlock;
            }
            else if (isBlock) {
                // Skip blocks
            }
            else if (markdownLines[i].trim() === "") {
                // Skip empty lines
            }
            else {
                markdownLines.splice(i, 0, generateAnchor(i));
            }
        }
        markdownLines.splice(0, 0, generateAnchor(i));
        markdownLines.push(generateAnchor(originalLinesCount - 1));
        const markedOptions = {
            baseUrl: this.props.oni.workspace.activeWorkspace,
            highlight(code, lang) {
                return code;
            },
        };
        const highlightsEnabled = this.props.oni.configuration.getValue("experimental.markdownPreview.syntaxHighlights");
        if (highlightsEnabled) {
            markedOptions.highlight = (code, lang) => {
                const languageExists = hljs.getLanguage(lang);
                const languageNotDefinedOrInvalid = typeof lang === "undefined" ||
                    (typeof languageExists === "undefined" && lang !== "nohighlight");
                if (languageNotDefinedOrInvalid) {
                    return hljs.highlightAuto(code).value;
                }
                if (lang === "nohighlight") {
                    return code;
                }
                return hljs.highlight(lang, code).value;
            };
        }
        marked.setOptions(markedOptions);
        return marked(markdownLines.join("\n"));
    }
    subscribe(editorEvent, eventCallback) {
        this._subscriptions.push(editorEvent.subscribe(eventCallback));
    }
    onBufferChanged(bufferInfo) {
        if (bufferInfo.buffer.language === "markdown") {
            this.previewBuffer(bufferInfo.buffer);
        }
    }
    onBufferScrolled(args) {
        if (this.props.oni.editors.activeEditor.activeBuffer.language !== "markdown") {
            return;
        }
        let anchor = null;
        for (let line = args.windowTopLine - 1; !anchor && line < args.bufferTotalLines; line++) {
            anchor = document.getElementById(generateScrollingAnchorId(line));
        }
        if (anchor) {
            anchor.scrollIntoView();
        }
    }
    previewBuffer(buffer) {
        buffer.getLines().then((lines) => {
            this.previewString(lines.join("\n"));
        });
    }
    previewString(str) {
        this.setState({ source: str });
    }
}
class MarkdownPreviewEditor {
    constructor(_oni) {
        this._oni = _oni;
        this._open = false;
        this._manuallyClosed = false;
        this._unrenderedContent = "";
        this._renderedContent = "";
        this._oni.editors.activeEditor.onBufferEnter.subscribe(args => this.onBufferEnter(args));
        this._oni.editors.activeEditor.onBufferLeave.subscribe(args => this.onBufferLeave(args));
    }
    isPaneOpen() {
        return this._open;
    }
    getUnrenderedContent() {
        return this._unrenderedContent;
    }
    getRenderedContent() {
        return this._renderedContent;
    }
    updateContent(unrendered, rendered) {
        this._unrenderedContent = unrendered;
        this._renderedContent = rendered;
    }
    toggle() {
        if (this._open) {
            this.close(true);
        }
        else {
            this.open();
        }
    }
    open() {
        if (!this._open) {
            this._open = true;
            this._manuallyClosed = false;
            const editorSplit = this._oni.windows.activeSplitHandle;
            // TODO: Update API
            this._split = this._oni.windows.createSplit("vertical", this);
            editorSplit.focus();
        }
    }
    close(manuallyClosed = false) {
        if (this._open) {
            this._open = false;
            this._manuallyClosed = manuallyClosed;
            this._split.close();
        }
    }
    render() {
        return React.createElement(MarkdownPreview, { oni: this._oni, instance: this });
    }
    onBufferEnter(bufferInfo) {
        if (bufferInfo.language === "markdown" && this._manuallyClosed === false) {
            this.open();
        }
    }
    onBufferLeave(bufferInfo) {
        this.close();
    }
}
function activate(oni) {
    if (!oni.configuration.getValue("experimental.markdownPreview.enabled", false)) {
        return;
    }
    const preview = new MarkdownPreviewEditor(oni);
    oni.commands.registerCommand(new Command("markdown.openPreview", "Open Markdown Preview", "Open the Markdown preview pane if it is not already opened", () => {
        preview.open();
    }));
    oni.commands.registerCommand(new Command("markdown.closePreview", "Close Markdown Preview", "Close the Markdown preview pane if it is not already closed", () => {
        preview.close(true);
    }));
    oni.commands.registerCommand(new Command("markdown.togglePreview", "Toggle Markdown Preview", "Open the Markdown preview pane if it is closed, otherwise open it", () => {
        preview.toggle();
    }));
    return preview;
}
exports.activate = activate;
class Command {
    constructor(command, name, detail, execute) {
        this.command = command;
        this.name = name;
        this.detail = detail;
        this.execute = execute;
    }
}
//# sourceMappingURL=index.js.map