"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const Oni = require("oni-api");
const oni_types_1 = require("oni-types");
const BookmarksSearch_1 = require("./BookmarksSearch");
const QuickOpenItem_1 = require("./QuickOpenItem");
class NullSearch {
    constructor() {
        this._onSearchResults = new oni_types_1.Event();
    }
    cancel() {
        return undefined;
    }
    changeQueryText(newText) {
        return undefined;
    }
    get onSearchResults() {
        return this._onSearchResults;
    }
}
class BufferLinesSearch {
    constructor(oni, lines) {
        this._onSearchResults = new oni_types_1.Event();
        let i = 0;
        this._items = lines.map(s => new QuickOpenItem_1.QuickOpenItem(String(++i), s, QuickOpenItem_1.QuickOpenType.bufferLine, oni.editors.activeEditor.activeBuffer.filePath, i, 0));
    }
    cancel() { }
    changeQueryText(newText) {
        this._onSearchResults.dispatch({
            items: this._items,
            isComplete: true,
        });
    }
    get onSearchResults() {
        return this._onSearchResults;
    }
}
class FilePathSearch {
    constructor(_oni) {
        this._oni = _oni;
        this._onSearchResults = new oni_types_1.Event();
        this._cached = false;
    }
    cancel() { }
    changeQueryText(newText) {
        if (this._cached) {
            return;
        }
        this._cached = true;
        const activeQuery = this._oni.search.findInPath({
            searchQuery: null,
            fileFilter: null,
            workspace: null,
        });
        activeQuery.onSearchResults.subscribe(result => {
            const items = result.items.map(i => FilePathSearch.toQuickOpenItem(i));
            this._onSearchResults.dispatch({
                items,
                isComplete: result.isComplete,
            });
        });
        activeQuery.start();
    }
    get onSearchResults() {
        return this._onSearchResults;
    }
    static toQuickOpenItem(item) {
        return new QuickOpenItem_1.QuickOpenItem(path.basename(item.fileName), path.dirname(item.fileName), QuickOpenItem_1.QuickOpenType.file, item.fileName, item.line, item.column);
    }
}
class FileContentSearch {
    constructor(_oni) {
        this._oni = _oni;
        this._onSearchResults = new oni_types_1.Event();
        this._activeQuery = this._oni.search.nullSearch;
    }
    cancel() {
        this._activeQuery.cancel();
        this._activeQuery = this._oni.search.nullSearch;
    }
    changeQueryText(newText) {
        this._activeQuery.cancel();
        if (!newText || newText.length < 1) {
            this._onSearchResults.dispatch({
                items: [],
                isComplete: true,
            });
            return;
        }
        const searchParams = {
            searchQuery: newText,
            fileFilter: null,
            workspace: null,
        };
        this._activeQuery = this._oni.search.findInFile(searchParams);
        this._activeQuery.onSearchResults.subscribe(result => {
            this._onSearchResults.dispatch({
                items: result.items.map(i => FileContentSearch.toQuickOpenItem(i)),
                isComplete: result.isComplete,
            });
        });
        this._activeQuery.start();
    }
    get onSearchResults() {
        return this._onSearchResults;
    }
    static toQuickOpenItem(item) {
        return new QuickOpenItem_1.QuickOpenItem(item.text, path.basename(item.fileName), QuickOpenItem_1.QuickOpenType.file, item.fileName, item.line, item.column);
    }
}
function getHome() {
    return process.env[process.platform === "win32" ? "USERPROFILE" : "HOME"];
}
class QuickOpen {
    constructor(_oni) {
        this._oni = _oni;
        this._searcher = new NullSearch();
        this._seenItems = new Set();
        this._itemsFound = [];
        this.setToQuickFix = () => __awaiter(this, void 0, void 0, function* () {
            this._oni.populateQuickFix(this._itemsFound.map(item => item.toQuickFixItem()));
            this._menu.hide();
        });
        this.openFileWithAltAction = () => {
            const mode = this._oni.configuration.getValue("editor.quickOpen.alternativeOpenMode");
            this.open(mode);
        };
        this.searchFileByContent = () => __awaiter(this, void 0, void 0, function* () {
            const filterName = "none"; // TODO: Use a filter like `regex` (needs a few adjustments)
            const searcher = new FileContentSearch(this._oni);
            yield this.search(searcher, filterName);
        });
        this.searchFileByPath = () => __awaiter(this, void 0, void 0, function* () {
            const filterName = this._oni.configuration.getValue("editor.quickOpen.filterStrategy", "vscode");
            const searchEngine = this.isInstallDirectoryOrHome()
                ? this._bookmarksSearch
                : new FilePathSearch(this._oni);
            yield this.search(searchEngine, filterName);
        });
        this.showBufferLines = () => __awaiter(this, void 0, void 0, function* () {
            const lines = yield this._oni.editors.activeEditor.activeBuffer.getLines();
            yield this.search(new BufferLinesSearch(this._oni, lines), "fuse");
        });
        this.showBookmarks = () => __awaiter(this, void 0, void 0, function* () {
            yield this.search(this._bookmarksSearch, "vscode");
        });
        this.search = (searcher, filterName) => __awaiter(this, void 0, void 0, function* () {
            this._searcher.cancel();
            const filterFunction = this._oni.filter.getByName(filterName);
            this._menu.setFilterFunction(filterFunction);
            this._searcher = searcher;
            searcher.onSearchResults.subscribe(result => {
                this._itemsFound = result.items;
                if (result.isComplete) {
                    this._menu.setLoading(false);
                }
                const pinned = (e) => this._seenItems.has(e.hash);
                const menuItems = result.items.map(e => e.toMenuItem(this._oni, pinned(e)));
                this._menu.setItems(menuItems);
            });
            this._menu.show();
            this._searcher.changeQueryText("");
        });
        this._menu = _oni.menu.create();
        this._menu.onHide.subscribe(() => {
            this._searcher.cancel();
        });
        this._bookmarksSearch = new BookmarksSearch_1.default(this._oni);
        this._menu.onFilterTextChanged.subscribe(newFilter => {
            if (this._menu.isOpen()) {
                const timeout = window.setTimeout(() => {
                    this._menu.setLoading(true);
                }, 200);
                this._searcher.changeQueryText(newFilter);
                if (timeout) {
                    window.clearTimeout(timeout);
                }
            }
            else {
                this._searcher.cancel();
            }
        });
        this._menu.onItemSelected.subscribe(() => {
            this.open(this.getDefaultOpenMode());
        });
    }
    isOpen() {
        return this._menu.isOpen();
    }
    open(mode) {
        return __awaiter(this, void 0, void 0, function* () {
            const selectedItem = this._menu.selectedItem;
            if (!selectedItem) {
                return;
            }
            const { activeWorkspace } = this._oni.workspace;
            const basePath = activeWorkspace ? [activeWorkspace] : [];
            const pathArgs = basePath.concat([selectedItem.metadata["path"]]);
            const fullPath = path.join(...pathArgs).replace("~", getHome());
            this._seenItems.add(selectedItem.metadata["hash"]);
            const qoType = QuickOpenItem_1.getTypeFromMenuItem(selectedItem);
            switch (qoType) {
                case QuickOpenItem_1.QuickOpenType.bookmarkHelp: {
                    this._oni.commands.executeCommand("oni.config.openConfigJs");
                    break;
                }
                case QuickOpenItem_1.QuickOpenType.folderHelp: {
                    this._oni.commands.executeCommand("workspace.openFolder");
                    break;
                }
                case QuickOpenItem_1.QuickOpenType.bufferLine: {
                    if (mode !== Oni.FileOpenMode.Edit) {
                        const openOptions = { openMode: mode };
                        const path = this._oni.editors.activeEditor.activeBuffer.filePath;
                        yield this._oni.editors.openFile(path, openOptions);
                    }
                    yield this._oni.editors.activeEditor.neovim.command(`${selectedItem.label}`);
                    break;
                }
                case QuickOpenItem_1.QuickOpenType.folder: {
                    this._oni.workspace.changeDirectory(fullPath);
                    break;
                }
                case QuickOpenItem_1.QuickOpenType.bookmark: {
                    // TODO: add an option to automatically open the quick open menu
                    // Also look at https://github.com/isaacs/node-glob, as a means
                    // of allowing users to specify globs for their bookmarks
                    yield this._bookmarksSearch.handleBookmark(selectedItem);
                    break;
                }
                case QuickOpenItem_1.QuickOpenType.file: {
                    const line = parseInt(selectedItem.metadata["line"], 10);
                    const column = parseInt(selectedItem.metadata["column"], 10);
                    const offsetedLine = line > 0 ? line - 1 : 0;
                    const offsetedColumn = column > 0 ? column - 1 : 0;
                    yield this._oni.editors.openFile(fullPath, { openMode: mode });
                    yield this._oni.editors.activeEditor.activeBuffer.setCursorPosition(offsetedLine, offsetedColumn);
                    break;
                }
                default: {
                    this._oni.log(`Error: unhandled QO type: ${qoType} (${QuickOpenItem_1.QuickOpenType[qoType]})`);
                    break;
                }
            }
            this._menu.hide();
        });
    }
    getDefaultOpenMode() {
        const legacy = this._oni.configuration.getValue("editor.quickOpen.defaultOpenMode", null);
        if (legacy) {
            return legacy;
        }
        return this._oni.configuration.getValue("quickOpen.defaultOpenMode", Oni.FileOpenMode.NewTab);
    }
    isInstallDirectoryOrHome() {
        // Offer to open folder/bookmark (Basically user hasn't opened a folder yet)
        const cwd = process.cwd();
        const installationDir = path.dirname(process.execPath);
        const homeDir = getHome();
        return cwd === installationDir || cwd === homeDir;
    }
}
exports.QuickOpen = QuickOpen;
//# sourceMappingURL=QuickOpen.js.map