"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const electron_1 = require("electron");
const PersistentSettings = require("electron-settings");
const installDevTools_1 = require("./installDevTools");
const Log = require("./Log");
const menu_1 = require("./menu");
const ProcessLifecycle_1 = require("./ProcessLifecycle");
const WindowManager_1 = require("./WindowManager");
global["getLogs"] = Log.getAllLogs;
const processArgs = process.argv || [];
const isAutomation = processArgs.find(f => f.indexOf("--test-type=webdriver") >= 0);
const isDevelopment = process.env.NODE_ENV === "development" || process.env.ONI_WEBPACK_LOAD === "1";
const isDebug = process.argv.filter(arg => arg.indexOf("--debug") >= 0).length > 0;
if (isAutomation) {
    Log.setVerbose(true);
    Log.info("Verbose flag set since running automation");
}
let windowState = {
    bounds: {
        x: null,
        y: null,
        height: 600,
        width: 800,
    },
    isMaximized: false,
};
function storeWindowState(main) {
    Log.info("storeWindowState called");
    if (!main) {
        return;
    }
    windowState.isMaximized = main.isMaximized();
    if (!windowState.isMaximized) {
        windowState.bounds = main.getBounds();
    }
    try {
        Log.info("Starting to persist settings...");
        PersistentSettings.set("_internal.windowState", windowState);
        Log.info("Completed setting persisted settings");
    }
    catch (e) {
        Log.info(`error setting window state: ${e.message}`);
    }
}
electron_1.ipcMain.on("focus-next-instance", () => {
    Log.info("focus-next-instance");
    focusNextInstance(1);
});
electron_1.ipcMain.on("focus-previous-instance", () => {
    Log.info("focus-previous-instance");
    focusNextInstance(-1);
});
electron_1.ipcMain.on("move-to-next-oni-instance", (event, direction) => {
    Log.info(`Attempting to swap to Oni instance on the ${direction}.`);
    WindowManager_1.moveToNextOniInstance(windows, direction);
});
electron_1.ipcMain.on("open-oni-window", () => {
    Log.info("opening window");
    createWindow([], process.cwd());
});
let windows = [];
const activeWindow = () => {
    return windows.filter(w => w.isFocused())[0] || null;
};
if (!isDevelopment && !isDebug && !isAutomation) {
    const currentOptions = {
        args: processArgs,
        workingDirectory: process.env["ONI_CWD"] || process.cwd(),
    };
    Log.info("Making single instance...");
    ProcessLifecycle_1.makeSingleInstance(currentOptions, options => {
        Log.info("Creating single instance");
        loadFileFromArguments(process.platform, options.args, options.workingDirectory);
    });
}
else {
    let argsToUse = processArgs;
    if (isAutomation) {
        Log.warn("Clearing arguments because running from automation!");
        argsToUse = [];
    }
    electron_1.app.on("ready", () => __awaiter(this, void 0, void 0, function* () {
        if (!isAutomation) {
            yield installDevTools_1.default();
        }
        loadFileFromArguments(process.platform, argsToUse, process.env.ONI_CWD || process.cwd());
    }));
}
function createWindow(commandLineArguments, workingDirectory, delayedEvent = null) {
    Log.info(`Creating window with arguments: ${commandLineArguments} and working directory: ${workingDirectory}`);
    const webPreferences = {
        blinkFeatures: "ResizeObserver,Accelerated2dCanvas,Canvas2dFixedRenderingMode",
    };
    const backgroundColor = PersistentSettings.get("_internal.lastBackgroundColor") || "#1E2127";
    try {
        const internalWindowState = PersistentSettings.get("_internal.windowState");
        if (internalWindowState &&
            (internalWindowState.bounds || internalWindowState.isMaximized)) {
            windowState = internalWindowState;
        }
    }
    catch (e) {
        Log.info(`error getting window state: ${e.message}`);
    }
    const rootPath = path.join(__dirname, "..", "..", "..");
    const iconImage = process.platform === "win32" ? "oni.ico" : "256x256.png";
    const iconPath = path.join(rootPath, "images", iconImage);
    const indexFileName = isDevelopment ? "index.dev.html" : "index.html";
    const indexPath = path.join(rootPath, indexFileName + "?react_perf");
    let currentWindow = new electron_1.BrowserWindow({
        icon: iconPath,
        webPreferences,
        backgroundColor,
        titleBarStyle: "hidden",
        x: windowState.bounds.x,
        y: windowState.bounds.y,
        height: windowState.bounds.height,
        width: windowState.bounds.width,
    });
    if (process.platform === "darwin") {
        currentWindow.webContents.on("devtools-opened", () => {
            currentWindow.webContents.devToolsWebContents.executeJavaScript(`
                window.addEventListener('keydown', function (e) {
                    if (e.keyCode === 65 && e.metaKey) {
                        document.execCommand('Select All');
                    } else if (e.keyCode === 67 && e.metaKey) {
                        document.execCommand('copy');
                    } else if (e.keyCode === 86 && e.metaKey) {
                        document.execCommand('paste');
                    } else if (e.keyCode === 87 && e.metaKey) {
                        window.close();
                    } else if (e.keyCode === 88 && e.metaKey) {
                        document.execCommand('cut');
                    }
                });`);
        });
    }
    if (windowState.isMaximized) {
        currentWindow.maximize();
    }
    updateMenu(currentWindow, false);
    currentWindow.webContents.on("did-finish-load", () => {
        Log.info("did-finish-load event received");
        currentWindow.webContents.send("init", {
            args: commandLineArguments,
            workingDirectory,
        });
    });
    electron_1.ipcMain.once("Oni.started", evt => {
        Log.info("Oni started");
        if (delayedEvent) {
            currentWindow.webContents.send(delayedEvent.evt, ...delayedEvent.cmd);
        }
    });
    electron_1.ipcMain.on("rebuild-menu", (_evt, loadInit) => {
        if (currentWindow) {
            updateMenu(currentWindow, loadInit);
        }
    });
    currentWindow.loadURL(`file://${indexPath}`);
    if (process.env.NODE_ENV === "development" || commandLineArguments.indexOf("--debug") >= 0) {
        currentWindow.webContents.openDevTools();
    }
    currentWindow.on("move", () => {
        storeWindowState(currentWindow);
    });
    currentWindow.on("resize", () => {
        storeWindowState(currentWindow);
    });
    currentWindow.on("close", () => {
        Log.info("close event...");
        storeWindowState(currentWindow);
        Log.info("...close event completed");
    });
    currentWindow.on("closed", () => {
        Log.info("closed event...");
        windows = windows.filter(m => m !== currentWindow);
        currentWindow = null;
        Log.info("...closed event completed");
    });
    if (!isDevelopment) {
        currentWindow.webContents.on("will-navigate", (event, url) => {
            event.preventDefault();
            currentWindow.webContents.send("execute-command", "browser.openUrl", url);
        });
    }
    windows.push(currentWindow);
    return currentWindow;
}
exports.createWindow = createWindow;
electron_1.app.on("open-file", (event, filePath) => {
    event.preventDefault();
    Log.info(`filePath to open: ${filePath}`);
    if (activeWindow()) {
        activeWindow().webContents.send("open-file", filePath);
    }
    else if (process.platform.includes("darwin")) {
        const argsToUse = [...process.argv.slice(1), filePath];
        if (electron_1.app.isReady()) {
            createWindow(argsToUse, process.cwd());
        }
        else {
            electron_1.app.on("ready", () => {
                createWindow(argsToUse, process.cwd());
            });
        }
    }
});
electron_1.app.on("window-all-closed", () => {
    Log.info("window-all-closed event");
    if (process.platform !== "darwin" || isAutomation) {
        Log.info("quitting app");
        electron_1.app.quit();
    }
});
electron_1.app.on("activate", () => {
    if (!windows.length) {
        createWindow([], process.cwd());
    }
    if (activeWindow()) {
        activeWindow().show();
    }
});
function updateMenu(browserWindow, loadInit) {
    const menu = menu_1.buildMenu(browserWindow, loadInit);
    if (process.platform === "darwin") {
        electron_1.Menu.setApplicationMenu(menu);
        const dockMenu = menu_1.buildDockMenu(browserWindow, loadInit);
        electron_1.app.dock.setMenu(dockMenu);
    }
    else {
        browserWindow.setMenu(menu);
    }
}
function focusNextInstance(direction) {
    const currentFocusedWindows = windows.filter(f => f.isFocused());
    if (currentFocusedWindows.length === 0) {
        Log.info("No window currently focused");
        return;
    }
    const currentFocusedWindow = currentFocusedWindows[0];
    const currentWindowIdx = windows.indexOf(currentFocusedWindow);
    let newFocusWindowIdx = (currentWindowIdx + direction) % windows.length;
    if (newFocusWindowIdx < 0) {
        newFocusWindowIdx = windows.length - 1;
    }
    Log.info(`Focusing index: ${newFocusWindowIdx}`);
    windows[newFocusWindowIdx].focus();
}
function loadFileFromArguments(platform, args, workingDirectory) {
    const localOni = "LOCAL_ONI";
    if (!process.env[localOni]) {
        createWindow(args.slice(1), workingDirectory);
    }
    else {
        createWindow(args.slice(2), workingDirectory);
    }
}
//# sourceMappingURL=main.js.map