/// <reference types="react" />
/// <reference types="node" />
import * as ChildProcess from "child_process";
import * as types from "vscode-languageserver-types";
import { IEvent } from "oni-types";
import * as Search from "./Search";
import * as Ui from "./Ui";
export { Search, Ui };
export declare type DisposeFunction = () => void;
export interface QuickFixEntry {
    filename: string;
    lnum: number;
    col: number;
    text: string;
}
/**
 * API surface area for registering and executing commands
 */
export declare namespace Commands {
    type CommandCallback = (args?: any) => any;
    type CommandEnabledCallback = () => boolean;
    interface ICommand {
        command: string;
        name: string;
        detail: string;
        enabled?: CommandEnabledCallback;
        messageSuccess?: string;
        messageFail?: string;
        execute: CommandCallback;
    }
    interface Api {
        registerCommand(command: ICommand): void;
        unregisterCommand(commandName: string): void;
        executeCommand(name: string, args?: any): boolean | void;
    }
}
export interface IToken {
    tokenName: string;
    range: types.Range;
}
export declare type ConfigurationValues = {
    [key: string]: any;
};
export interface Configuration {
    onConfigurationChanged: IEvent<ConfigurationValues>;
    getValue<T>(configValue: string, defaultValue?: T): T;
    setValues(configurationValues: ConfigurationValues): void;
}
export declare namespace Notifications {
    interface Notification {
        onClick: IEvent<void>;
        onClose: IEvent<void>;
        setContents(title: string, detail: string): void;
        setExpiration(expirationTimeInMilliseconds: number): void;
        show(): void;
        hide(): void;
    }
    interface Api {
        enable(): void;
        disable(): void;
        createItem(): Notification;
    }
}
export declare namespace Overlays {
    /**
     * An overlay is a full-screen UI element,
     * that renders above the editor layer.
     */
    interface Overlay {
        hide(): void;
        show(): void;
        setContents(element: JSX.Element): void;
    }
    interface Api {
        createItem(): Overlay;
    }
}
export declare namespace Workspace {
    interface Api {
        activeWorkspace: string;
        onDirectoryChanged: IEvent<string>;
        changeDirectory(newWorkspaceDirectory: string): void;
        applyEdits(edits: types.WorkspaceEdit): Promise<void>;
    }
}
export declare namespace Snippets {
    /**
     * Metadata describing a snippet
     */
    interface Snippet {
        prefix: string;
        body: string;
        description: string;
    }
    /**
     * `SnippetProvider` is a strategy for providing snippets for a particular language
     *
     * This can range from loading from a file/network to ad-hoc configuration in user config.
     */
    interface SnippetProvider {
        getSnippets(language: string): Promise<Snippet[]>;
    }
    /**
     * `SnippetManager` is the main entry point for integrating with snippets.
     */
    interface SnippetManager {
        isSnippetActive: boolean;
        insertSnippet(snippet: string): Promise<void>;
        cancel(): Promise<void>;
        nextPlaceholder(): Promise<void>;
        previousPlaceholder(): Promise<void>;
        getSnippetsForLanguage(language: string): Promise<Snippet[]>;
        registerSnippetProvider(snippetProvider: SnippetProvider): void;
    }
}
export declare type Direction = "left" | "right" | "down" | "up";
export declare type SplitDirection = "horizontal" | "vertical";
export interface WindowSplitHandle {
    id: string;
    close(): void;
    show(): void;
    hide(): void;
    focus(): void;
    setSize(size: number): void;
}
export interface IWindowManager {
    onUnhandledMove: IEvent<Direction>;
    activeSplitHandle: WindowSplitHandle;
    createSplit(direction: Direction | SplitDirection, split: IWindowSplit): WindowSplitHandle;
    focusSplit(splitId: string): void;
    move(direction: Direction): void;
}
export interface IWindowSplit {
    render(): JSX.Element;
}
export declare enum FileOpenMode {
    Edit = 0,
    VerticalSplit = 1,
    HorizontalSplit = 2,
    NewTab = 3,
    ExistingTab = 4,
}
export interface FileOpenOptions {
    openMode: FileOpenMode;
}
export declare const DefaultFileOpenOptions: FileOpenOptions;
export interface EditorManager {
    /**
     * An array of all available `Editor` instances
     */
    allEditors: Editor[];
    /**
     * A proxy that _always_ points to the active `Editor`. This means
     * you can avoid always needing to hook/unhook events when the
     * active `Editor` changes by hooking events on `anyEditor`
     */
    anyEditor: Editor;
    /**
     * The currently active `Editor` instance
     */
    activeEditor: Editor;
    /**
     * Event that is dispatched when the active `Editor` changes,
     * for example, when focus moves from one `Editor` to another.
     */
    onActiveEditorChanged: IEvent<Editor>;
    openFile(filePath: string, options?: FileOpenOptions): Promise<Buffer>;
}
/**
 * Input API entry point
 */
export declare namespace Input {
    type InputAction = () => void;
    /**
     * Often, you want a keybinding to be available in certain modes, but not other modes.
     * Vim handles this by exposing several variations of its `map` command (like `nmap`, `imap`, `noremap`, etc)
     *
     * In Oni, the filter function determines whether or not a binding ins enabled. It's simply a function
     * that returns `true` or `false`.
     */
    type InputFilter = () => boolean;
    interface InputManager {
        /**
         * Bind a key or set of keys to an _action_
         *
         * The _action_ can either be a JavaScript callback, or a _command_ string.
         *
         * To see available command strings, check out our [KeyBindings file](https://github.com/onivim/oni/blob/master/browser/src/Input/KeyBindings.ts)
         *
         * `filterFunction` is an optional third argument. If it returns `true`, the input
         * binding is enabled. This is helpful, for example, to enable key bindings only
         * in certain conditions.
         *
         * An example usage of the filter function might be:
         * ```
         * const isNormalMode = () => oni.editors.activeEditor.mode === "normal"
         * const isVisualMode = () => oni.editors.activeEditor.mode === "visual"
         * oni.input.bind("<esc>", () => alert("Escape pressed in normal mode!"), isNormalMode)
         * oni.input.bind("<esc>", () => alert("Escape pressed in visual mode!"), isVisualMode)
         * ```
         *
         * In this case, the `isNormalMode` and `isVisualMode` functions are used as _filters_. When the `<esc>` key is pressed,
         * all bindings for `<esc>` are evaluated, and the first one with a passing filter (or no filter) is used.
         */
        bind(keyChord: string | string[], actionOrCommand: InputAction | string, filterFunction?: InputFilter): void;
        hasBinding(keyChord: string): boolean;
        /**
         * `unbind` removes an keybindings present on a key.
         *
         * `unbind` is useful for removing default functionality. When your configuration is loaded, the default
         * keybindings provided by Oni are applied, but you can always opt-out of these in order to restore
         * default functionality.
         *
         * For example, if I would prefer to use the [CtrlP](https://github.com/ctrlpvim/ctrlp.vim) plugin, you could [install that plugin](https://github.com/onivim/oni/wiki/Plugins#installing-a-vim-plugin),
         * and then use `oni.input.unbind("<c-p>")` to prevent Oni from opening the bundled file opener. When there are no keybindings present, the input sequence is passed to the active editor (usually Neovim) to handle.
         */
        unbind(keyChord: string | string[]): void;
        unbindAll(): void;
    }
}
export interface IPluginManager {
    getPlugin(name: string): any;
    loaded: boolean;
}
/**
 * Automation API definition
 *
 * The Automation API provides a set of utilities
 * geared towards testing Oni's functionality.
 */
export declare namespace Automation {
    type WaitConditionChecker = () => boolean;
    interface Api {
        sendKeys(key: string): void;
        sleep(time: number): Promise<void>;
        waitFor(condition: WaitConditionChecker, timeout?: number): Promise<void>;
        waitForEditors(): Promise<void>;
    }
}
export interface NeovimEditorCapability {
    callFunction(functionName: string, args: any[]): Promise<any>;
    input(keys: string): Promise<void>;
    eval(expression: string): Promise<any>;
    command(command: string): Promise<void>;
}
/**
 * A BufferLayer is a UI overlay presented over a buffer.
 *
 * This allows for all manners of custom rendering - whether it is simply
 * overlaying tokens or showing adorners, or completely overriding the
 * rendering layer of the buffer.
 */
export interface BufferLayer {
    /**
     * Unique id for the buffer layer. This must be globally unique and is used to reference the layer.
     */
    id: string;
    /**
     * `friendlyName` is used when showing the layer int eh UI
     */
    friendlyName?: string;
    /**
     * `render` returns a custom UI element. Note that this may be called multiple times per buffer,
     * if there are multiple windows with the same buffer.
     */
    render(context: BufferLayerRenderContext): JSX.Element;
}
/**
 * BufferLayerRenderContext
 */
export interface BufferLayerRenderContext {
    isActive: boolean;
    windowId: number;
    /**
     * The width of a character cell, in pixels
     */
    fontPixelWidth: number;
    /**
     * The height of a character cell, in pixels
     */
    fontPixelHeight: number;
    /**
     * The line number of the first line in view
     */
    topBufferLine: number;
    bottomBufferLine: number;
    /**
     * Array of the visible lines, with visibleLines[0] representing the line at `topBufferLine`,
     * with length of `bottomBufferLine - topBufferLine`.
     */
    visibleLines: string[];
    /**
     * Function that takes a _buffer_ position (line, character) and converts it to a
     * _pixel space_ position
     */
    bufferToPixel: Coordinates.BufferToPixel;
    /**
     * Function that takes a _buffer_ position (line, character) and converts it to a
     * _screen cell_ position.
     */
    bufferToScreen: Coordinates.BufferToScreen;
    /**
     * Function that takes a _screen cell_ position and converts it to a _pixel_ position.
     */
    screenToPixel: Coordinates.ScreenToPixel;
    /**
     * Dimensions of the buffer window, in cells.
     */
    dimensions: Shapes.Rectangle;
}
export declare type InputCallbackFunction = (input: string) => Promise<void>;
export declare type EditorTextOptions = {
    /**
     * Whether or not spaces should be inserted for tab characters
     */
    insertSpacesForTab: boolean;
    /**
     * Size of tabs, in characters
     */
    tabSize: number;
};
export interface Editor {
    mode: string;
    activeBuffer: Buffer;
    init(filesToOpen: string[]): void;
    /**
     * Helper function to queue / block input while a long-running process
     * is occurring.
     *
     * This is important for API calls that may require multiple steps while
     * the user is typing, for example, auto-closing pairs or snippets.
     *
     * This takes a function that returns a promise, as well as an input
     * argument so that input can be entered exclusively by the block.
     *
     * Use sparingly as this may cause a visible delay in typing.
     */
    blockInput(atomicInputFunction: (inputCallback: InputCallbackFunction) => Promise<void>): void;
    openFile(file: string, openOptions?: FileOpenOptions): Promise<Buffer>;
    getBuffers(): Array<Buffer | InactiveBuffer>;
    setTextOptions(textOptions: EditorTextOptions): Promise<void>;
    setSelection(selectionRange: types.Range): Promise<void>;
    render(): JSX.Element;
    onBufferEnter: IEvent<EditorBufferEventArgs>;
    onBufferLeave: IEvent<EditorBufferEventArgs>;
    onBufferChanged: IEvent<EditorBufferChangedEventArgs>;
    onBufferScrolled: IEvent<EditorBufferScrolledEventArgs>;
    onBufferSaved: IEvent<EditorBufferEventArgs>;
    onCursorMoved: IEvent<Cursor>;
    onModeChanged: IEvent<Vim.Mode>;
    neovim?: NeovimEditorCapability;
}
export interface EditorBufferChangedEventArgs {
    buffer: Buffer;
    contentChanges: types.TextDocumentContentChangeEvent[];
}
export interface EditorBufferScrolledEventArgs {
    bufferTotalLines: number;
    windowTopLine: number;
    windowBottomLine: number;
}
export interface Buffer {
    id: string;
    language: string;
    filePath: string;
    cursor: Cursor;
    version: number;
    modified: boolean;
    lineCount: number;
    addLayer(layer: BufferLayer): void;
    removeLayer(layer: BufferLayer): void;
    applyTextEdits(edit: types.TextEdit | types.TextEdit[]): Promise<void>;
    getLines(start?: number, end?: number): Promise<string[]>;
    getTokenAt(line: number, column: number): Promise<IToken>;
    getSelectionRange(): Promise<types.Range>;
    setLines(start: number, end: number, lines: string[]): Promise<void>;
    setCursorPosition(line: number, column: number): Promise<void>;
    setLanguage(lang: string): Promise<void>;
    getCursorPosition(): Promise<types.Position>;
    handleInput(key: string): boolean;
    setScratchBuffer(): Promise<void>;
}
export interface InactiveBuffer {
    id: string;
    language: string;
    filePath: string;
    version: number;
    modified: boolean;
    lineCount: number;
}
export interface Cursor {
    line: number;
    column: number;
}
export interface EditorBufferEventArgs {
    language: string;
    filePath: string;
}
export interface Log {
    verbose(msg: string): void;
    info(msg: string): void;
    disableVerboseLogging(): void;
    enableVerboseLogging(): void;
}
/**
 * The `recorder` API enables the taking of screenshot or recording a video.
 *
 * The output defaults to the `os.tmpdir()` directory, but can be overridden
 * by the `recorder.outputPath` option.
 */
export interface Recorder {
    /**
     * Returns `true` if a recording is in progress, `false` otherwise
     */
    isRecording: boolean;
    /**
     * Start recording a video.
     *
     * Videos are recorded in `.webm` format.
     */
    startRecording(): void;
    /**
     * Stop recording a video.
     *
     * Optionally specify the destination `fileName`.
     */
    stopRecording(fileName?: string): Promise<void>;
    /**
     * Take a screenshot.
     *
     * Optionally specify the destination `fileName.
     */
    takeScreenshot(fileName?: string): void;
}
export interface StatusBar {
    getItem(globalId?: string): StatusBarItem;
    createItem(alignment: number, globalId?: string): StatusBarItem;
}
export interface Process {
    execNodeScript(scriptPath: string, args?: string[], options?: ChildProcess.ExecOptions, callback?: (err: any, stdout: string, stderr: string) => void): Promise<ChildProcess.ChildProcess>;
    spawnNodeScript(scriptPath: string, args?: string[], options?: ChildProcess.SpawnOptions): Promise<ChildProcess.ChildProcess>;
    spawnProcess(startCommand: string, args?: string[], options?: ChildProcess.SpawnOptions): Promise<ChildProcess.ChildProcess>;
}
export interface StatusBarItem {
    show(): void;
    hide(): void;
    setContents(element: JSX.Element): void;
    dispose(): void;
}
export declare namespace Vim {
    type Mode = "normal" | "visual" | "insert";
}
export declare namespace Coordinates {
    interface PixelSpacePoint {
        pixelX: number;
        pixelY: number;
    }
    interface ScreenSpacePoint {
        screenX: number;
        screenY: number;
    }
    type BufferToScreen = (position: types.Position) => ScreenSpacePoint | null;
    type ScreenToPixel = (screenPoint: ScreenSpacePoint) => PixelSpacePoint | null;
    type BufferToPixel = (position: types.Position) => PixelSpacePoint | null;
}
export declare namespace Shapes {
    interface Rectangle {
        x: number;
        y: number;
        width: number;
        height: number;
    }
    namespace Rectangle {
        function create(x: number, y: number, width: number, height: number): Rectangle;
    }
}
export declare namespace ToolTip {
    enum OpenDirection {
        Up = 1,
        Down = 2,
    }
    interface ToolTipOptions {
        position?: Coordinates.PixelSpacePoint;
        openDirection: OpenDirection;
        padding?: string;
        onDismiss?: () => void;
    }
}
export declare namespace Sidebar {
    interface Pane {
        id: string;
        title: string;
        [method: string]: any;
    }
    interface SidebarEntry {
        hasNotification: boolean;
        icon: string;
        id: string;
        pane: Sidebar.Pane;
    }
    interface Api {
        activeEntryId: string;
        isFocused: boolean;
        entries: Sidebar.SidebarEntry[];
    }
}
export declare namespace Menu {
    interface MenuOption {
        /**
         * Optional font-awesome icon
         */
        icon?: string;
        label: string;
        detail?: string;
        metadata?: {
            [id: string]: string;
        };
        /**
         * A pinned option is always shown first in the menu,
         * before unpinned items
         */
        pinned?: boolean;
        /**
         * Score used to sort the results.
         */
        score?: number;
    }
    interface IMenuOptionWithHighlights extends Menu.MenuOption {
        labelHighlights: number[];
        detailHighlights: number[];
    }
    type IMenuFilter = (options: any[], searchString: string) => IMenuOptionWithHighlights[];
    type filterFunc = (items: Menu.MenuOption[], searchString: string) => IMenuOptionWithHighlights[];
    interface IMenuFilters {
        getDefault(): IMenuFilter;
        getByName(name: string): IMenuFilter;
    }
    interface MenuInstance {
        onHide: IEvent<void>;
        onItemSelected: IEvent<any>;
        onSelectedItemChanged: IEvent<Menu.MenuOption>;
        onFilterTextChanged: IEvent<string>;
        selectedItem: MenuOption;
        setLoading(isLoading: boolean): void;
        setItems(items: MenuOption[]): void;
        isOpen(): boolean;
        show(): void;
        hide(): void;
        setFilterFunction(filterFn: filterFunc): void;
    }
    interface Api {
        create: () => MenuInstance;
        closeActiveMenu: () => void;
        isMenuOpen: () => boolean;
        nextMenuItem: () => void;
        previousMenuItem: () => void;
        selectMenuItem: (index?: number) => Menu.MenuOption;
    }
}
export interface IColors {
    getColor(colorName: string): string | null;
}
export declare namespace Plugin {
    namespace Diagnostics {
        interface Api {
            setErrors(key: string, fileName: string, errors: types.Diagnostic[]): void;
        }
    }
    interface Api {
        automation: Automation.Api;
        colors: IColors;
        commands: Commands.Api;
        configuration: Configuration;
        contextMenu: any;
        diagnostics: Diagnostics.Api;
        editors: EditorManager;
        filter: Menu.IMenuFilters;
        input: Input.InputManager;
        language: any;
        log: any;
        notifications: Notifications.Api;
        overlays: Overlays.Api;
        plugins: IPluginManager;
        search: Search.ISearch;
        sidebar: Sidebar.Api;
        ui: Ui.IUi;
        menu: Menu.Api;
        process: Process;
        recorder: Recorder;
        snippets: Snippets.SnippetManager;
        statusBar: StatusBar;
        windows: IWindowManager;
        workspace: Workspace.Api;
        populateQuickFix(entries: QuickFixEntry[]): void;
    }
}
