"use strict";
/**
 * Git.ts
 *
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const oni_types_1 = require("oni-types");
const GitP = require("simple-git/promise");
const VCS = require("./vcs"); // TODO: import from oni-api
class GitVersionControlProvider {
    constructor(_oni, _git = GitP) {
        this._oni = _oni;
        this._git = _git;
        this._name = "git";
        this._onPluginActivated = new oni_types_1.Event("Oni::VCSPluginActivated");
        this._onPluginDeactivated = new oni_types_1.Event("Oni::VCSPluginDeactivated");
        this._onBranchChange = new oni_types_1.Event("Oni::VCSBranchChanged");
        this._onStagedFilesChanged = new oni_types_1.Event("Oni::VCSStagedFilesChanged");
        this._onFileStatusChanged = new oni_types_1.Event("Oni::VCSFilesStatusChanged");
        this._isActivated = false;
        this.getStatus = () => __awaiter(this, void 0, void 0, function* () {
            try {
                const status = yield this._git(this._projectRoot).status();
                const { modified, staged } = this._getModifiedAndStaged(status.files);
                return {
                    staged,
                    modified,
                    ahead: status.ahead,
                    behind: status.behind,
                    created: status.created,
                    deleted: status.deleted,
                    currentBranch: status.current,
                    conflicted: status.conflicted,
                    untracked: status.not_added,
                    remoteTrackingBranch: status.tracking,
                };
            }
            catch (e) {
                this._handleVCSError(e, "get current status");
            }
        });
        this.getBlame = (args) => __awaiter(this, void 0, void 0, function* () {
            try {
                const cmd = [
                    "--no-pager",
                    "blame",
                    "--porcelain",
                    args.file,
                    "--show-number",
                    "-L",
                    `${args.lineOne},${args.lineTwo}`,
                ];
                const raw = yield this._git(this._projectRoot).raw(cmd);
                return this._formatRawBlame(raw);
            }
            catch (e) {
                this._oni.log.warn(e);
                return null;
            }
        });
        this.unstage = (files) => __awaiter(this, void 0, void 0, function* () {
            const flags = ["HEAD", ...files];
            try {
                yield this._git(this._projectRoot).reset(flags);
                const changed = files.map(path => ({ path, status: VCS.Statuses.modified }));
                this._onFileStatusChanged.dispatch(changed);
            }
            catch (e) {
                this._handleVCSError(e, "unstage the file");
            }
        });
        this.uncommit = (sha) => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this._git(this._projectRoot).reset(["--soft", "HEAD^"]);
                this._onFileStatusChanged.dispatch();
            }
            catch (e) {
                this._handleVCSError(e, "undo most recent commit");
            }
        });
        this.getLogs = (file) => __awaiter(this, void 0, void 0, function* () {
            try {
                // n - represents the number of logs to get alternative is to use ["--max-count=25"]
                const options = { file, n: 25 };
                return this._git(this._projectRoot).log(options);
            }
            catch (e) {
                this._handleVCSError(e, "get logs");
            }
        });
        this.getDiff = () => __awaiter(this, void 0, void 0, function* () {
            try {
                return this._git(this._projectRoot).diffSummary();
            }
            catch (e) {
                this._handleVCSError(e, "get current status");
            }
        });
        this.commitFiles = (message, files) => __awaiter(this, void 0, void 0, function* () {
            try {
                const commit = yield this._git(this._projectRoot).commit(message, files);
                const changed = (files || []).map(file => ({
                    path: file,
                    status: VCS.Statuses.committed,
                }));
                this._onFileStatusChanged.dispatch(changed);
                return commit;
            }
            catch (e) {
                this._handleVCSError(e, "commit files");
            }
        });
        this.stageFile = (file) => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this._git(this._projectRoot).add(file);
                this._onStagedFilesChanged.dispatch(file);
                this._onFileStatusChanged.dispatch([{ path: file, status: VCS.Statuses.staged }]);
            }
            catch (e) {
                this._handleVCSError(e, `add ${file}`);
            }
        });
        this.fetchBranchFromRemote = ({ branch, currentDir, remote = null, }) => __awaiter(this, void 0, void 0, function* () {
            try {
                return this._git(this._projectRoot).fetch(remote, branch);
            }
            catch (e) {
                this._handleVCSError(e, "to fetch branch");
            }
        });
        this.getLocalBranches = () => __awaiter(this, void 0, void 0, function* () {
            try {
                return this._git(this._projectRoot).branchLocal();
            }
            catch (e) {
                this._handleVCSError(e, "get local branches");
            }
        });
        this.getBranch = () => __awaiter(this, void 0, void 0, function* () {
            try {
                const status = yield this._git(this._projectRoot).status();
                return status.current;
            }
            catch (e) {
                this._handleVCSError(e, "get current status");
            }
        });
        this._isStaged = (file) => {
            const GitPIndicators = ["M", "A"];
            return GitPIndicators.some(status => file.index.includes(status));
        };
        this._projectRoot = this._oni.workspace.activeWorkspace;
        this._oni.workspace.onDirectoryChanged.subscribe(workspace => {
            this._projectRoot = workspace;
        });
    }
    get onBranchChanged() {
        return this._onBranchChange;
    }
    get onFileStatusChanged() {
        return this._onFileStatusChanged;
    }
    get onStagedFilesChanged() {
        return this._onStagedFilesChanged;
    }
    get onPluginActivated() {
        return this._onPluginActivated;
    }
    get onPluginDeactivated() {
        return this._onPluginDeactivated;
    }
    get isActivated() {
        return this._isActivated;
    }
    get name() {
        return this._name;
    }
    activate() {
        this._isActivated = true;
        this._onPluginActivated.dispatch();
    }
    deactivate() {
        this._isActivated = false;
        this._onPluginDeactivated.dispatch();
    }
    canHandleWorkspace(dir) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return this._git(this._projectRoot)
                    .silent()
                    .checkIsRepo();
            }
            catch (e) {
                this._oni.log.warn(`Git provider was unable to check if this directory is a repository because ${e.message}`);
                return false;
            }
        });
    }
    getRoot() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return this._git(this._projectRoot).revparse(["--show-toplevel"]);
            }
            catch (e) {
                this._handleVCSError(e, `find vcs root`);
            }
        });
    }
    changeBranch(targetBranch) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this._git(this._projectRoot).checkout(targetBranch);
                this._onBranchChange.dispatch(targetBranch);
            }
            catch (e) {
                this._handleVCSError(e, "change branch");
            }
        });
    }
    _handleVCSError(error, attemptedAction) {
        this._oni.log.warn(error);
        throw new Error(`[Oni Git Provider]: Unable to ${attemptedAction} because: ${error.message}`);
    }
    _formatRawBlame(rawOutput) {
        if (!rawOutput) {
            return null;
        }
        const firstSpace = (str) => str.indexOf(" ");
        const blameArray = rawOutput.split("\n");
        const formatted = blameArray
            .map(line => [line.substr(0, firstSpace(line)), line.substr(firstSpace(line) + 1)])
            .reduce((acc, [key, value], index) => {
            const formattedKey = key.replace("-", "_");
            if (!index && value) {
                acc.hash = formattedKey;
                if (value) {
                    const [originalLine, finalLine, numberOfLines] = value.split(" ");
                    acc.line = {
                        originalLine,
                        finalLine,
                        numberOfLines,
                    };
                }
                return acc;
            }
            else if (!key) {
                return acc;
            }
            acc[formattedKey] = value;
            return acc;
        }, {});
        return formatted;
    }
    _getModifiedAndStaged(files) {
        return files.reduce((acc, file) => {
            if (file.working_dir === "M") {
                acc.modified.push(file.path);
            }
            else if (this._isStaged(file)) {
                acc.staged.push(file.path);
            }
            return acc;
        }, { modified: [], staged: [] });
    }
}
exports.GitVersionControlProvider = GitVersionControlProvider;
exports.activate = (oni) => __awaiter(this, void 0, void 0, function* () {
    const provider = new GitVersionControlProvider(oni);
    yield oni.services.vcs.registerProvider(provider);
    return provider;
});
//# sourceMappingURL=index.js.map